import { ItemBox } from './layouts/Layout.js';
export declare class VirtualRepeater<Item, Child extends Element, Key> {
    /**
     * Invoked when a new (not recycled) child needs to be displayed.
     * Set by createElement.
     */
    private _createElementFn;
    /**
     * Invoked on new (including recycled) children after the range changes.
     * Set by updateElement.
     */
    private _updateElementFn;
    /**
     * Invoked on outgoing children after the range changes. Set by recycleElement.
     */
    private _recycleElementFn;
    /**
     * Used for generating a key for storing references to children. Set by elementKey.
     */
    private _elementKeyFn;
    /**
     * Items to render. Set by items.
     */
    private _items;
    /**
     * Total number of items to render. Set by totalItems.
     */
    private _totalItems;
    /**
     * Flag for asynchnronous reset requests. Resetting clears and reinserts
     * children in the range.
     */
    private _needsReset;
    /**
     * Flag for asynchnronous remeasure requests. Signals that all children
     * should be remeasured.
     */
    private _needsRemeasure;
    /**
     * Used to keep track of children that can be cleaned up after a range update.
     */
    private _active;
    private _prevActive;
    /**
     * Used for recycling purposes.
     */
    private _keyToChild;
    private _childToKey;
    /**
     * Used to keep track of measures by index.
     */
    private _indexToMeasure;
    private __incremental;
    /**
     * Used for tracking range changes.
     */
    private _prevNum;
    /**
     * Invoked at the end of each render cycle: children in the range are
     * measured, and their dimensions passed to this callback. Use it to layout
     * children as needed.
     */
    protected _measureCallback: (sizes: {
        [key: number]: ItemBox;
    }) => void;
    /**
     * Number of children in the range. Set by num.
     * TODO @straversi: Consider renaming this. count? visibleElements?
     */
    protected _num: number;
    /**
     * Index of the first child in the range, not necessarily the first visible child.
     * TODO @straversi: Consider renaming these.
     */
    protected _first: number;
    /**
     * Index of the last child in the range.
     */
    protected _last: number;
    /**
     * Previous first rendered index. Used to avoid unnecessary updates.
     */
    protected _prevFirst: number;
    /**
     * Previous last rendered index. Used to avoid unnecessary updates.
     */
    protected _prevLast: number;
    /**
     * Flag for asynchnronous render requests. Renders can be requested several
     * times before a render actually happens.
     */
    protected _pendingRender: any;
    /**
     * Containing element. Set by container.
     */
    protected _container: Element | ShadowRoot;
    /**
     * Children in the rendered order.
     */
    protected _ordered: Array<Child>;
    constructor(config: any);
    /**
     * The parent of all child nodes to be rendered.
     */
    container: Element | ShadowRoot;
    /**
     * Invoked to create a child.
     * Override or set directly to control element creation.
     */
    createElement: (item: Item, index: number) => Child;
    /**
     * Invoked to update a child.
     * Override or set directly to control element updates.
     */
    updateElement: (child: Child, item: Item, index: number) => void;
    /**
     * Invoked in place of removing a child from the DOM, so that it can be reused.
     * Override or set directly to prepare children for reuse.
     */
    recycleElement: (child: Child, index: number) => void;
    /**
     * Invoked to generate a key for an element.
     * Override or set directly to control how keys are generated for children.
     */
    elementKey: (index: number) => Key;
    /**
     * The index of the first child in the range.
     */
    first: number;
    /**
     * The number of children in the range.
     */
    num: number;
    /**
     * An array of data to be used to render child nodes.
     */
    items: Array<Item>;
    /**
     * The total number of items, regardless of the range, that can be rendered
     * as child nodes.
     */
    totalItems: number;
    /**
     * TODO @straversi: Document this.
     */
    _incremental: boolean;
    /**
     * Invoke to request that all elements in the range be measured.
     */
    requestRemeasure(): void;
    protected _shouldRender(): boolean;
    /**
     * Render at the next opportunity.
     */
    protected _scheduleRender(): Promise<void>;
    /**
     * Invoke to request that all elements in the range be updated.
     */
    private requestReset;
    /**
     * Returns those children that are about to be displayed and that require to
     * be positioned. If reset or remeasure has been triggered, all children are
     * returned.
     */
    private readonly _toMeasure;
    /**
     * Measures each child bounds and builds a map of index/bounds to be passed
     * to the `_measureCallback`
     */
    private _measureChildren;
    /**
     * Render items within the current range to the DOM.
     */
    protected _render(): Promise<void>;
    /**
     * Invoked after DOM is updated, and before it gets measured.
     */
    protected _didRender(): void;
    /**
     * Unassigns any children at indices lower than the start of the current
     * range.
     */
    private _discardHead;
    /**
     * Unassigns any children at indices higher than the end of the current
     * range.
     */
    private _discardTail;
    /**
     * Assigns and inserts non-existent children from the current range with
     * indices lower than the start of the previous range.
     */
    private _addHead;
    /**
     * Assigns and appends non-existent children from the current range with
     * indices higher than the end of the previous range.
     */
    private _addTail;
    /**
     * Re-insert and update children in the given range.
     */
    private _reset;
    /**
     * Instantiates, tracks, and returns the child at idx.
     * Prevents cleanup of children that already exist.
     * Returns the new child at idx.
     */
    private _assignChild;
    /**
     * Removes the child at idx, recycling it if possible.
     */
    private _unassignChild;
    /**
     * Returns the node for the first child in the current range, if the node is
     * in the DOM.
     * TODO: Is this the right name?
     */
    readonly _firstChild: Node;
    /**
     * Overridable abstractions for child manipulation
     */
    /**
     * Return the node for child.
     * Override if child !== child's node.
     */
    _node(child: Child): Node;
    /**
     * Returns the next node sibling of a child node.
     */
    _nextSibling(child: Child): Node;
    /**
     * Inserts child before referenceNode in the container.
     * Override to control child insertion.
     */
    _insertBefore(child: Child, referenceNode: Node): void;
    /**
     * Removes child from the DOM.
     * TODO @straversi: this will not work (_node should be called on child), but
     * is not currently being used or tested.
     */
    _removeChild(child: Child): void;
    /**
     * Returns whether the child's node is a child of the container
     * element.
     */
    _childIsAttached(child: Child): boolean;
    /**
     * Sets the display style of the given node to 'none'.
     */
    _hideChild(child: Child): void;
    /**
     * Sets the display style of the given node to null.
     */
    _showChild(child: Child): void;
    /**
     * Returns the width, height, and margins of the given child.
     * Override if child !== child's node.
     */
    _measureChild(child: Child): ItemBox;
}
//# sourceMappingURL=VirtualRepeater.d.ts.map